@@ -137,8 +137,8 @@ class Ignnition_model:
         # add the file with any additional function, if any
         if 'additional_functions_file' in self.CONFIG:
             additional_path = self.__process_path(self.CONFIG['additional_functions_file'])
-            sys.path.insert(1, additional_path)
-            self.module = __import__(additional_path.split('/')[-1][0:-3])
+            sys.path.insert(1, os.path.join(additional_path, os.pardir))
+            self.module = __import__(os.path.basename(additional_path)[0:-3])
 
         self.model_info = self.__create_model()
         self.generator = Generator()
@@ -175,14 +175,37 @@ class Ignnition_model:
 
         return total_loss
 
+    def __denormalized_metric(self, metric, metric_name, denorm_func):
+        def denorm_metric(y_true, y_pred):
+            output_name = self.model_info.get_output_info()
+            denorm_y_true = tf.py_function(func=denorm_func, inp=[y_true, output_name], Tout=tf.float32)
+            denorm_y_pred = tf.py_function(func=denorm_func, inp=[y_pred, output_name], Tout=tf.float32)
+            return metric(denorm_y_true, denorm_y_pred)
+
+        denorm_metric.__name__ = 'denorm_{}'.format(metric_name)
+
+        return denorm_metric
+
     def __get_keras_metrics(self):
         metric_names = self.CONFIG['metrics']
         metrics = []
+
+        try:
+            denorm_func = getattr(self.module, 'denormalization')
+        except:
+            denorm_func = None
+
         for name in metric_names:
             if hasattr(tf.keras.metrics, name):
-                metrics.append(getattr(tf.keras.metrics, name)())
+                metric = getattr(tf.keras.metrics, name)()
+                metrics.append(metric)
+                if denorm_func is not None:
+                    metrics.append(self.__denormalized_metric(metric, metric.name, denorm_func))
             elif hasattr(self.module, name):
-                metrics.append(getattr(self.module, name))
+                metric = getattr(self.module, name)
+                metrics.append(metric)
+                if denorm_func is not None:
+                    metrics.append(self.__denormalized_metric(metric, name, denorm_func))
 
         return metrics
 
@@ -345,7 +368,7 @@ class Ignnition_model:
             adj_names = self.model_info.get_adjacency_info()
             interleave_list = self.model_info.get_interleave_tensors()
             interleave_sources = self.model_info.get_interleave_sources()
-            output_name = self.model_info.get_output_info()
+            output_names = self.model_info.get_output_info()
             additional_input = self.model_info.get_additional_input_names()
             unique_additional_input = [a for a in additional_input if a not in feature_list]
             entity_names = self.model_info.get_entity_names()
@@ -353,7 +376,7 @@ class Ignnition_model:
             feature_names = []
 
             for a in unique_additional_input:
-                types[a] = tf.int64
+                types[a] = tf.float32
                 shapes[a] = tf.TensorShape(None)
 
             for f_name in feature_list:
@@ -386,7 +409,7 @@ class Ignnition_model:
                 if data_samples is None:
                     ds = tf.data.Dataset.from_generator(
                         lambda: self.generator.generate_from_dataset(filenames, entity_names, feature_names,
-                                                                     output_name,  # adjacency_info,
+                                                                     output_names, adj_names,
                                                                      interleave_list, unique_additional_input, training,
                                                                      shuffle),
                         output_types=(types, tf.float32),
@@ -396,7 +419,7 @@ class Ignnition_model:
                     data_samples = [json.dumps(t) for t in data_samples]
                     ds = tf.data.Dataset.from_generator(
                         lambda: self.generator.generate_from_array(data_samples, entity_names, feature_names,
-                                                                   output_name,  # adjacency_info,
+                                                                   output_names, adj_names,
                                                                    interleave_list,
                                                                    unique_additional_input, training, shuffle),
                         output_types=(types, tf.float32),
@@ -406,7 +429,7 @@ class Ignnition_model:
                 if data_samples is None:
                     ds = tf.data.Dataset.from_generator(
                         lambda: self.generator.generate_from_dataset(filenames, entity_names, feature_names,
-                                                                     output_name,  # adjacency_info,
+                                                                     output_names, adj_names,
                                                                      interleave_list, unique_additional_input, training,
                                                                      shuffle),
                         output_types=(types),
@@ -416,7 +439,7 @@ class Ignnition_model:
                     data_samples = [json.dumps(t) for t in data_samples]
                     ds = tf.data.Dataset.from_generator(
                         lambda: self.generator.generate_from_array(data_samples, entity_names, feature_names,
-                                                                   output_name,  # adjacency_info,
+                                                                   output_names, adj_names,
                                                                    interleave_list,
                                                                    unique_additional_input, training, shuffle),
                         output_types=(types),
@@ -427,13 +450,13 @@ class Ignnition_model:
                 if batch_norm is None:
                     if training:
                         ds = ds.map(
-                            lambda x, y: self.__global_normalization(x, feature_list, output_name, y),
+                            lambda x, y: self.__global_normalization(x, feature_list, output_names, y),
                             num_parallel_calls=tf.data.experimental.AUTOTUNE)
                         ds = ds.prefetch(tf.data.experimental.AUTOTUNE)
 
                     else:
                         ds = ds.map(
-                            lambda x: self.__global_normalization(x, feature_list, output_name),
+                            lambda x: self.__global_normalization(x, feature_list, output_names),
                             num_parallel_calls=tf.data.experimental.AUTOTUNE)
                         ds = iter(ds)
 
@@ -491,10 +514,10 @@ class Ignnition_model:
             Input dictionary necessary to initialize all the dimensions
         """
 
-        checkpoint_path = self.CONFIG.get('warm_start_path', '')
+        checkpoint_path = self.CONFIG.get('load_model_path', '')
         if os.path.isfile(checkpoint_path):
             print("Restoring from", checkpoint_path)
-            # in this case we need to initialize the weights to be able to use a warm-start checkpoint
+            # in this case we need to initialize the weights to be able to use a load_model checkpoint
 
             sample_it = self.__input_fn_generator(training=False,
                                                   data_samples=[sample])
@@ -529,7 +552,7 @@ class Ignnition_model:
             if sample_paths == []:
                 print_failure("No dataset found. Please make sure the paths of the datasets are correct.")
             else:
-                sample_path = sample_paths[0] # choose one single file to extract the dimensions
+                sample_path = sample_paths[0]  # choose one single file to extract the dimensions
 
             if '.tar.gz' in sample_path:
                 try:
@@ -544,12 +567,16 @@ class Ignnition_model:
                 file_samples = open(sample_path, 'r')
 
             try:
-                file_samples.read(1)
+                ch1 = file_samples.read(1)
+                if ch1 != '[':
+                    print_failure(
+                        "Error because the dataset files must be an array of json objects, and not single json objects")
+
                 aux = stream_read_json(file_samples)
                 sample = next(aux)  # read one single example #json.load(file_samples)[0]  # one single sample
 
             except:
-                print_failure('Failed to read the data file ' + sample)
+                print_failure('Failed to read the data file ' + sample_path)
 
             # Now that we have the sample, we can process the dimensions
             dimensions = {}  # for each key, we have a tuple of (length, num_elements)
@@ -588,7 +615,7 @@ class Ignnition_model:
             graph_attrs = list(G.graph.keys())
             for g in graph_attrs:
                 feature = G.graph[g]
-                dimensions[g] = len(feature)
+                dimensions[g] = len(feature) if isinstance(feature, list) else 1
 
             return dimensions, sample
 
@@ -670,16 +697,23 @@ class Ignnition_model:
             Indicates if there should be verbosity in the prints of the terminal or not.
         """
 
-        prediction_path = None
+        # Extract the path to the prediction dataset, if any
+        data_path = None
+        if prediction_samples is None:
+            try:
+                data_path = self.__process_path(self.CONFIG['predict_dataset'])
+            except:
+                print_failure(
+                    'Make sure to either pass an array of samples or to define in the train_options.yaml the path to the predict dataset')
+
+        # create the GNN model --and load the previous checkpoint-- if it does not exist already
         if not hasattr(self, 'gnn_model'):
             if prediction_samples is None:  # look for the dataset path
                 try:
-                    prediction_path = self.__process_path(self.CONFIG['predict_dataset'])
-                    self.__create_gnn(path=prediction_path, verbose=verbose)
+                    self.__create_gnn(path=data_path, verbose=verbose)
                 except:
                     print_failure(
                         'Make sure to either pass an array of samples or to define in the train_options.yaml the path to the prediction dataset')
-
             else:
                 self.__create_gnn(samples=prediction_samples, verbose=verbose)
 
@@ -688,7 +722,7 @@ class Ignnition_model:
             print_header(
                 'Starting to make the predictions...\n---------------------------------------------------------\n')
 
-        sample_it = self.__input_fn_generator(prediction_path, training=False, data_samples=prediction_samples,
+        sample_it = self.__input_fn_generator(data_path, training=False, data_samples=prediction_samples,
                                               iterator=True)
         all_predictions = []
         try:
@@ -758,11 +792,19 @@ class Ignnition_model:
             Indicates if there should be verbosity in the prints of the terminal or not.
         """
 
+        # Extract the path to the validation dataset, if any
+        data_path = None
+        if evaluation_samples is None:
+            try:
+                data_path = self.__process_path(self.CONFIG['validation_dataset'])
+            except:
+                print_failure(
+                    'Make sure to either pass an array of samples or to define in the train_options.yaml the path to the validation dataset')
+
         # Generate the model if it doesn't exist
         if not hasattr(self, 'gnn_model'):
             if evaluation_samples is None:  # look for the dataset path
-                val_path = self.__process_path(self.CONFIG['validation_dataset'])
-                self.__create_gnn(path=val_path)
+                self.__create_gnn(path=data_path)
             else:
                 self.__create_gnn(samples=evaluation_samples)
 
@@ -770,15 +812,6 @@ class Ignnition_model:
             print()
             print_header('Starting to make evaluations...\n---------------------------------------------------------\n')
 
-        if evaluation_samples is None:
-            try:
-                data_path = self.CONFIG['validation_dataset']
-            except:
-                print_failure(
-                    'Make sure to either pass an array of samples or to define in the train_options.yaml the path to the validation dataset')
-        else:
-            data_path = None
-
         sample_it = self.__input_fn_generator(data_path, training=True, data_samples=evaluation_samples, iterator=True)
 
         all_metrics = []
